# -*- coding:utf-8 -*-

from __future__ import division, print_function, absolute_import

import codecs

import numpy as np

import nagisa_utils as utils

OOV = utils.OOV
PAD = utils.PAD


def embedding_loader(fn_embedding, word2id, init_range=0.1):
    # load a pre-trained embedding file (word2vec format)
    word2vec = {}
    with codecs.open(fn_embedding, 'r', encoding='utf_8_sig') as f:
        for i, line in enumerate(f):
            line = line.rstrip()
            if i == 0:
                dim_word = int(line.split(' ')[1])
            else:
                line = line.split(' ')
                word = line[0].lower()
                if word in word2id:
                    vec = np.asarray(list(map(float, line[1:])))
                    word2vec[word] = vec

    embs = []
    for word, idx in sorted(word2id.items(), key=lambda x:x[-1]):
        if word.lower() in word2vec:
            embs.append(word2vec[word.lower()])
        else:
            unk = np.random.uniform(-init_range, init_range, dim_word)
            embs.append(unk)
    embs = np.asarray(embs)
    return embs, dim_word


def update_dict(key, dictionary):
    if key in dictionary:
        dictionary[key] += 1
    else:
        dictionary[key] = 1
    return dictionary


def cut_by_threshold(dictionary, oov, pad, threshold=2):
    token2id = {oov:0, pad:1}
    for token, freq in reversed(sorted(dictionary.items(), key=lambda x:x[1])):
        if freq >= threshold:
            token2id[token] = len(token2id)
    return token2id


def create_vocabs_from_trainset(trainset, threshold=2,
                                fn_dictionary=None, save_vocabs=True,
                                fn_vocabs=None, delimiter='\t',
                                newline='EOS', oov=OOV, pad=PAD):
    # Creat a word-to-POStags dictionary.
    word2postags = {}
    if fn_dictionary is not None:
        with codecs.open(fn_dictionary, 'r', encoding='utf_8_sig') as texts:
            for text in texts:
                text = utils.utf8rstrip(text)
                word, postag = text.split('\t')
                word = utils.normalize(word)
                # lower setting: 1
                word = word.lower()
                if word in word2postags:
                    word2postags[word].append(postag)
                else:
                    word2postags[word] = [postag]

    # Creat a word-to-index dictionary and a index-to-word dictionary.
    dictionary = {oov:0, pad:1}
    for word in word2postags.keys():
        dictionary[word] = len(dictionary)
    id2word = {i:w for w, i in dictionary.items()}

    # Creat a unigram-to-index dictionary, a bigram-to-index dictionary.
    # Reconstruct a word-to-index dictionary.
    words   = []
    uni2id  = {}
    bi2id   = {}
    word2id = {}
    pos2id  = {oov:0}
    with codecs.open(trainset, 'r', encoding='utf_8_sig') as texts:
        for text in texts:
            text = utils.utf8rstrip(text)
            if text == newline:
                sent = ''.join(words)
                unis = utils.get_unigram(sent)
                for uni in unis:
                    uni2id = update_dict(uni, uni2id)

                bis = utils.get_bigram(sent)
                for bi in bis:
                    bi2id = update_dict(bi, bi2id)

                words_at_i = utils.get_words_starting_at_i(sent, dictionary)
                words_at_i += utils.get_words_ending_at_i(sent, dictionary)
                for words in words_at_i:
                    for wid in words:
                        word    = id2word[wid]
                        word2id = update_dict(word, word2id)
                words = []

            else:
                word, pos = text.split(delimiter)
                word = utils.normalize(word)
                word = word.replace(' ', '　')
                # lower setting: 2
                word = word.lower()
                words.append(word)
                word2id = update_dict(word, word2id)
                pos2id  = update_dict(pos, pos2id)

    # Cut keys by frequency threshold.
    uni2id  = cut_by_threshold(uni2id,  oov, pad, threshold)
    bi2id   = cut_by_threshold(bi2id,   oov, pad, threshold)
    word2id = cut_by_threshold(word2id, oov, pad, threshold)

    # Creat a POStag-to-index dictionary.
    pos2id = {k:i for i,k in enumerate(pos2id.keys())}
    word2postags = {k:[pos2id[p] if p in pos2id else pos2id[oov] for p in list(set(v))]\
                    for k,v in word2postags.items()}

    vocabs = [uni2id, bi2id, word2id, pos2id, word2postags]
    if save_vocabs is True:
        utils.dump_data(vocabs, fn_vocabs)

    return vocabs


class from_file(object):
    def __init__(self, filename, window_size, vocabs,
                 newline='EOS', delimiter='\t'):
        self.words    = []
        self.ws_data  = []
        self.pos_data = []
        self.filename = filename
        uni2id, bi2id, word2id, pos2id, word2postags = vocabs

        if u'名詞' in pos2id:
            self.use_noun_heuristic = True
        else:
            self.use_noun_heuristic = False

        with codecs.open(filename, 'r', encoding='utf_8_sig') as texts:
            wids  = [] # Word index
            cids  = [] # Character index
            pids  = [] # POStag index
            words = [] # Original Words
            ptags = [] # Original POStags
            for text in texts:
                text = utils.utf8rstrip(text)
                if text == newline:
                    sent = ''.join(words)
                    segmented_sent = ' '.join(words)
                    tags = utils.make_tags_as_bmes(segmented_sent)

                    feats = utils.feature_extraction(text=sent,
                                                     uni2id=uni2id,
                                                     bi2id=bi2id,
                                                     dictionary=word2id,
                                                     window_size=window_size)

                    self.words.append(words)
                    self.ws_data.append([feats, tags])
                    self.pos_data.append([[cids, wids, ptags], pids])
                    # reset index lists
                    wids  = []
                    cids  = []
                    pids  = []
                    words = []
                    ptags = []

                else:
                    word, pos = text.split(delimiter)
                    word = utils.normalize(word)
                    word = word.replace(' ', '　')
                    # lower setting: 3
                    word = word.lower()

                    if word in word2postags:
                        w2p = word2postags[word]
                    else:
                        w2p = [0] # OOV
                    if self.use_noun_heuristic is True:
                        if word.isalnum() is True:
                            if w2p == [0]:
                                w2p = [pos2id[u'名詞']]
                            else:
                                w2p.append(pos2id[u'名詞'])

                    w2p = list(set(w2p))
                    ptags.append(w2p)

                    words.append(word)
                    if word in word2id:
                        wids.append(word2id[word])
                    else:
                        wids.append(word2id[OOV])

                    if pos in pos2id:
                        pids.append(pos2id[pos])
                    else:
                        pids.append(pos2id[OOV])

                    char_seq = []
                    for char in word:
                        if char in uni2id:
                            char_seq.append(uni2id[char])
                        else:
                            char_seq.append(uni2id[OOV])
                    cids.append(char_seq)



# -*- coding:utf-8 -*-

from __future__ import division, print_function, absolute_import

import os
import re
import sys

import nagisa_utils as utils
import nagisa.model as model

base = os.path.dirname(os.path.abspath(__file__))
sys.path.append(base)


class Tagger(object):
    """
    This class has a word segmentation function and a POS-tagging function for Japanese.
    """

    def __init__(self, vocabs=None, params=None, hp=None, single_word_list=None):
        if vocabs is None:
            vocabs = base + '/data/nagisa_v001.dict'
        if params is None:
            params = base + '/data/nagisa_v001.model'
        if hp is None:
            hp = base + '/data/nagisa_v001.hp'

        # Load vocaburary files
        vocabs = utils.load_data(vocabs)
        self._uni2id, self._bi2id, self._word2id, self._pos2id, self._word2postags = vocabs
        self._id2pos = {v:k for k, v in self._pos2id.items()}
        self.id2pos  = self._id2pos
        self.postags = [postag for postag in self._pos2id.keys()]
        # Load a hyper-parameter file
        self._hp = utils.load_data(hp)
        # Construct a word segmentation model and a pos tagging model
        self._model = model.Model(self._hp, params)

        # If a word is included in the single_word_list,
        # it is recognized as a single word forcibly.
        self.pattern = None
        if single_word_list:
            single_word_list = [utils.preprocess(w) for w in single_word_list if len(w) > 1]
            single_word_list = [w.replace('(', '\(').replace(')', '\)')
                                for w in single_word_list]
            single_word_list = sorted(single_word_list, key=lambda x:-len(x))
            if len(single_word_list) > 0:
                self.pattern = re.compile('|'.join(single_word_list))

        # If use_noun_heuristic is True, nouns are more lilely to appear.
        if u'名詞' in self._pos2id:
            self.use_noun_heuristic = True
        else:
            self.use_noun_heuristic = False


    def wakati(self, text, lower=False):
        """Word segmentation function. Return the segmented words.

        args:
            - text (str): An input sentence.
            - lower (bool): If lower is True, all uppercase characters in a list \
                            of the words are converted into lowercase characters.

        return:
            - words (list): A list of the words.
        """
        text = utils.preprocess(text)
        lower_text = text.lower()
        feats = utils.feature_extraction(text=lower_text,
                                         uni2id=self._uni2id,
                                         bi2id=self._bi2id,
                                         dictionary=self._word2id,
                                         window_size=self._hp['WINDOW_SIZE'])
        obs  = self._model.encode_ws(feats)
        obs  = [ob.npvalue() for ob in obs]
        tags = utils.np_viterbi(self._model.trans_array, obs)

        # A word can be recognized as a single word forcibly.
        if self.pattern:
            for match in self.pattern.finditer(text):
                span = match.span()
                span_s = span[0]
                span_e = span[1]

                if (span_e - span_s) == 1:
                    tags[span_s:span_e] = [3]
                else:
                    tags[span_s:span_e] = [0]+[1]*((span_e-span_s)-2)+[2]

                if span_s != 0:
                    previous_tag = tags[span_s-1]
                    if previous_tag == 0:   # 0 is BEGIN tag
                        tags[span_s-1] = 3  # 3 is SINGLE tag
                    elif previous_tag == 1: # 1 is MIDDEL tag
                        tags[span_s-1] = 2  # 2 is END tag

                if span_e != len(text):
                    next_tag = tags[span_e]
                    if next_tag == 1:    # 1 is MIDDEL tag
                        tags[span_e] = 0 # 0 is BEGIN tag
                    elif next_tag == 2:  # 2 is END tag
                        tags[span_e] = 3 # 3 is SINGLE tag

        if lower is True:
            words = utils.segmenter_for_bmes(lower_text, tags)
        else:
            words = utils.segmenter_for_bmes(text, tags)
        return words


    def _postagging(self, words, lower=False):
        if lower is True:
            words = [w.lower() for w in words]

        wids = utils.conv_tokens_to_ids(words, self._word2id)
        cids = [utils.conv_tokens_to_ids([c for c in w], self._uni2id) for w in words]
        tids = []
        for w in words:
            if w in self._word2postags:
                w2p = self._word2postags[w]
            else:
                w2p = [0]
            if self.use_noun_heuristic is True:
                if w.isalnum() is True:
                    if w2p == [0]:
                        w2p = [self._pos2id[u'名詞']]
                    else:
                        w2p.append(self._pos2id[u'名詞'])
            w2p = list(set(w2p))
            tids.append(w2p)

        X = [cids, wids, tids]
        postags = [self._id2pos[pid] for pid in self._model.POStagging(X)]
        return postags


    def postagging(self, words, lower=False):
        """ Return the words with POS-tags of the given words.

        args:
            - words (list): Input words.
            - lower (bool): If lower is True, all uppercase characters in a list \
                            of the words are converted into lowercase characters.
        return:
            - object : The object of the words with POS-tags.
        """
        return self.decode(words, lower)


    def decode(self, words, lower=False):
        """ Return the words with tags of the given words.

        args:
            - words (list): Input words.
            - lower (bool, optional): If lower is True, all uppercase characters in a list \
                            of the words are converted into lowercase characters.
        return:
            - object : The object of the words with tags.
        """
        if not type(words) == list:
            raise AssertionError("Please input a list of words.")
        words = [utils.preprocess_without_rstrip(w) if w == " " or w == "　"
                 else utils.preprocess(w) for w in words]
        postags = self._postagging(words, lower)
        return postags


    def tagging(self, text, lower=False):
        """ Return the words with POS-tags of the given sentence.

        args:
            - text (str): An input sentence.
            - lower (bool): If lower is True, all uppercase characters in a list \
                            of the words are converted into lowercase characters.
        return:
            - object : The object of the words with POS-tags.
        """
        return self._Token(text, lower, self.wakati, self._postagging)


    def filter(self, text, lower=False, filter_postags=None):
        """Return the filtered words with POS-tags of the given sentence.

        args:
            - text (str): An input sentence.
            - lower (bool): If lower is True, all uppercase characters in a list \
                            of the words are converted into lowercase characters.
            - filter_postags (list): Filtering the word with the POS-tag in \
                                     filter_postags from a text.

        return:
            - object : The object of the words with POS-tags.
        """
        if filter_postags is None:
            filter_postags = []

        words   = []
        postags = []
        tokens  = self.tagging(text, lower)
        for word, postag in zip(tokens.words, tokens.postags):
            if not postag in filter_postags:
                words.append(word)
                postags.append(postag)
        return self._Token(text, lower, self.wakati, self._postagging,
                           _words=words, _postags=postags)


    def extract(self, text, lower=False, extract_postags=None):
        """Return the extracted words with POS-tags of the given sentence.

        args:
            - text (str): An input sentence.
            - lower (bool): If lower is True, all uppercase characters in a list \
                            of the words are converted into lowercase characters.
            - filter_postags (list): Extracting the word with the POS-tag in \
                                     filter_postags from a text.

        return:
            - object : The object of the words with POS-tags.

        """
        if extract_postags is None:
            extract_postags = []

        words   = []
        postags = []
        tokens  = self.tagging(text, lower)
        for word, postag in zip(tokens.words, tokens.postags):
            if postag in extract_postags:
                words.append(word)
                postags.append(postag)
        return self._Token(text, lower, self.wakati, self._postagging,
                           _words=words, _postags=postags)


    class _Token(object):
        def __init__(self, text, lower, wakati, postagging, _words=None, _postags=None):
            self.text = text
            self.__lower = lower
            self.__words = _words
            self.__postags = _postags
            self.__wakati = wakati
            self.__postagging = postagging

        @property
        def words(self):
            if self.__words is None:
                self.__words = self.__wakati(self.text, self.__lower)
            return self.__words

        @property
        def postags(self):
            if self.__postags is None:
                self.__postags = self.__postagging(self.words, self.__lower)
            return self.__postags

        def __str__(self):
            return ' '.join([w+'/'+p for w, p in zip(self.words, self.postags)])



# -*- coding:utf-8 -*-

from __future__ import division, print_function, absolute_import

import re
import sys
import gzip
import codecs
import numpy as np
import unicodedata

from six.moves import cPickle

reload(sys)
if sys.version_info.major == 2:
    sys.setdefaultencoding('utf-8')

cdef unicode __OOV = u'oov'
cdef unicode __PAD = u'pad'

OOV = __OOV
PAD = __PAD

_hiragana = re.compile(u'[\u3040-\u309F]')
_katakana = re.compile(u'[\u30A1-\u30FA]')
_kanji    = re.compile(u'[\u4e00-\u9fa5]')
_alpha    = re.compile(u'[a-zA-Z]')
_numeric  = re.compile(u'[0-9]')


cpdef unicode utf8rstrip(text):
    if type(text) != unicode:
        return unicode(text.rstrip(), 'utf-8')
    else:
        return text.rstrip()


cpdef unicode normalize(unicode text):
    return unicodedata.normalize('NFKC', text)


cpdef unicode preprocess(text):
    text = utf8rstrip(text)
    text = normalize(text)
    text = text.replace('İ', 'I')
    text = text.replace(' ', '　')
    return text


cpdef unicode preprocess_without_rstrip(text):
    if type(text) != unicode:
        text = unicode(text, 'utf-8')
    text = normalize(text)
    text = text.replace('İ', 'I')
    text = text.replace(' ', '　')
    return text


cpdef list get_unigram(unicode text):
    cdef unicode uni
    return [uni for uni in text]


cpdef list get_bigram(unicode text):
    cdef:
        int i
        int length_text = len(text)
        unicode end_symbol = u'<E>'
    return [text[i]+end_symbol if i == length_text-1 else text[i:i+2]
            for i in range(length_text)]


cpdef int get_chartype(unicode character):
    if _hiragana.search(character):
        return 0
    elif _katakana.search(character):
        return 1
    elif _kanji.search(character):
        return 2
    elif _alpha.search(unicodedata.normalize('NFKC', character)):
        return 3
    elif _numeric.search(unicodedata.normalize('NFKC', character)):
        return 4
    else:
        return 5


cpdef list get_words_starting_at_i(unicode text, dict dictionary):
    cdef:
        int i
        int j
        int length_text = len(text)
        list subwords
        list words_starting_at_i = []
        unicode sub

    for i in range(length_text):
        subwords = []
        for j in range(i, min(i+8, length_text)):
            sub = text[i:j+1]
            if sub in dictionary:
                subwords.append(dictionary[sub])
        if len(subwords) == 0:
            subwords.append(dictionary[__OOV])
        words_starting_at_i.append(subwords)
    return words_starting_at_i


cpdef list get_words_ending_at_i(unicode text, dict dictionary):
    cdef:
        int i
        int j
        int length_text = len(text)
        list subwords
        list words_ending_at_i = []

    text = text[::-1]
    for i in range(length_text):
        subwords = []
        for j in range(i, min(i+8, length_text)):
            sub = text[i:j+1][::-1]
            if sub in dictionary:
                subwords.append(dictionary[sub])
        if len(subwords) == 0:
            subwords.append(dictionary[__OOV])
        words_ending_at_i.append(subwords)
    return words_ending_at_i[::-1]


cpdef list conv_tokens_to_ids(list words, dict word2id):
    cdef unicode word
    return [word2id[word] if word in word2id else word2id[__OOV] for word in words]


cpdef list context_window(list l, int win, int pad_id=1):
    cdef:
        int length_l = len(l)

    assert (win % 2) == 1
    assert win >=1
    lpadded = int(win/2) * [pad_id] + l + int(win/2) * [pad_id]
    out = [lpadded[i:i+win] for i in range(length_l)]
    assert len(out) == len(l)
    return out


cpdef list feature_extraction(unicode text, dict uni2id, dict bi2id,
                              dict dictionary, int window_size):
    # character-level features
    unigrams = get_unigram(text)
    bigrams = get_bigram(text)
    uids = context_window(conv_tokens_to_ids(unigrams, uni2id), window_size)
    bids = context_window(conv_tokens_to_ids(bigrams, bi2id), window_size)
    cids = context_window([get_chartype(uni) for uni in unigrams], window_size, pad_id=6)

    # word-level features
    wids_s = get_words_starting_at_i(text, dictionary)
    wids_e = get_words_ending_at_i(text, dictionary)

    features = [uids, bids, cids, wids_s, wids_e]
    return features


cpdef dict load_dictionary(dict_path):
    cdef dict word_dict = {__OOV:0, __PAD:1}
    with open(dict_path, 'r') as words:
        for word in words:
            word = utf8rstrip(word)
            if not word in word_dict:
                word_dict[word] = len(word_dict)
    return word_dict


cpdef list make_tags_as_bmes(unicode text):
    cdef:
        int i
        int len_word
        list tags = []
        list words = text.split(u' ')
    for word in words:
        len_word = len(word)
        if len_word < 2:
            tags += [3]
        elif len_word == 2:
            tags += [0, 2]
        elif len_word > 2:
            for i in range(len_word):
                if i == 0:
                    tags += [0]
                elif i == len_word-1:
                    tags += [2]
                else:
                    tags += [1]
    assert len(''.join(words)) == len(tags)
    return tags


cpdef list segmenter_for_bmes(unicode chars, list tags):
    cdef:
        int tag
        list words = []
        unicode partical_word = u''
        unicode character

    assert len(chars) == len(tags)
    for character, tag in zip(chars, tags):
        if tag == 3:
            words.append(character)
        elif tag == 2:
            partical_word += character
            words.append(partical_word)
            partical_word = u''
        else:
            partical_word += character
    return words


cpdef dump_data(data, fn):
    with gzip.open(fn, 'wb') as gf:
        cPickle.dump(data, gf, protocol=2)
        gf.close()


cpdef load_data(fn):
    with gzip.open(fn, 'rb') as gf:
        return cPickle.load(gf)


cpdef list np_viterbi(trans, observations):
    cdef:
        int idx, best_tag_id
        list bptrs_t, vvars_t, backpointer, indice, best_path

    for_expr = np.array([-1e10]*6)
    for_expr[4] = 0 # sp_s = 4
    indice = [0,1,2,3,4,5]
    backpointer = []

    for obs in observations:
        bptrs_t = []
        vvars_t = []
        for idx in indice:
            next_tag_expr = for_expr+trans[idx]
            best_tag_id = np.argmax(next_tag_expr)
            bptrs_t.append(best_tag_id)
            vvars_t.append(next_tag_expr[best_tag_id])
        for_expr = np.array(vvars_t) + obs
        backpointer.append(bptrs_t)

    terminal_expr = for_expr + trans[5] # sp_e = 5
    best_tag_id = np.argmax(terminal_expr)
    best_path = [best_tag_id]

    for bptrs_t in reversed(backpointer):
        best_tag_id = bptrs_t[best_tag_id]
        best_path.append(best_tag_id)

    best_path.pop()
    best_path.reverse()
    return best_path


cpdef load_file(filename, delimiter='\t', newline='EOS'):
    cdef:
        list X, Y, words, tags
        unicode word, tag

    X = []
    Y = []
    words = []
    tags = []

    with codecs.open(filename, 'r', encoding='utf_8_sig') as f:
        for line in f:
            line = utf8rstrip(line)

            if line == newline:
                if not len(words) == len(tags):
                    raise AssertionError("len(words) != len(tags)")

                X.append(words)
                Y.append(tags)
                words = []
                tags = []
            else:
                line = line.split(delimiter)
                word = " ".join(line[:-1])
                tag = line[-1]
                words.append(word)
                tags.append(tag)

    return X, Y



# -*- coding:utf-8 -*-

from __future__ import division, print_function, absolute_import

import time
import random
import logging
from collections import OrderedDict


import model
import prepro
import mecab_system_eval
import nagisa_utils as utils

from tagger import Tagger

logging.basicConfig(level=logging.INFO, format='%(message)s')
logger = logging.getLogger(__name__)


def fit(train_file, dev_file, test_file, model_name,
        dict_file=None, emb_file=None, delimiter='\t', newline='EOS',
        layers=1, min_count=2, decay=1, epoch=10, window_size=3,
        dim_uni=32, dim_bi=16, dim_word=16, dim_ctype=8, dim_tagemb=16,
        dim_hidden=100, learning_rate=0.1, dropout_rate=0.3, seed=1234):

    """Train a joint word segmentation and sequence labeling (e.g, POS-tagging, NER) model.

    args:
        - train_file (str): Path to a train file.
        - dev_file (str): Path to a development file for early stopping.
        - test_file (str):  Path to a test file for evaluation.
        - model_name (str): Output model filename.
        - dict_file (str, optional): Path to a dictionary file.
        - emb_file (str, optional): Path to a pre-trained embedding file (word2vec format).
        - delimiter (str, optional): Separate word and tag in each line by 'delimiter'.
        - newline (str, optional):  Separate lines in the file by 'newline'.
        - layers (int, optional): RNN Layer size.
        - min_count (int, optional): Ignores all words with total frequency lower than this.
        - decay (int, optional): Learning rate decay.
        - epoch (int, optional): Epoch size.
        - window_size (int, optional): Window size of the context characters for word segmentation.
        - dim_uni (int, optional): Dimensionality of the char-unigram vectors.
        - dim_bi (int, optional): Dimensionality of the char-bigram vectors.
        - dim_word (int, optional): Dimensionality of the word vectors.
        - dim_ctype (int, optional): Dimensionality of the character-type vectors.
        - dim_tagemb (int, optional): Dimensionality of the tag vectors.
        - dim_hidden (int, optional): Dimensionality of the BiLSTM's hidden layer.
        - learning_rate (float, optional): Learning rate of SGD.
        - dropout_rate (float, optional): Dropout rate of the input vector for BiLSTMs.
        - seed (int, optional): Random seed.

    return:
        - Nothing. After finish training, however,
          save the three model files (*.vocabs, *.params, *.hp) in the current directory.

    """

    random.seed(seed)

    hp = OrderedDict({
          'LAYERS':layers,
          'THRESHOLD':min_count,
          'DECAY':decay,
          'EPOCH':epoch,
          'WINDOW_SIZE':window_size,
          'DIM_UNI':dim_uni,
          'DIM_BI':dim_bi,
          'DIM_WORD':dim_word,
          'DIM_CTYPE':dim_ctype,
          'DIM_TAGEMB':dim_tagemb,
          'DIM_HIDDEN':dim_hidden,
          'LEARNING_RATE':learning_rate,
          'DROPOUT_RATE':dropout_rate,
          'SEED': seed,

          'TRAINSET':train_file,
          'TESTSET':test_file,
          'DEVSET':dev_file,
          'DICTIONARY':dict_file,
          'EMBEDDING':emb_file,

          'HYPERPARAMS':model_name+'.hp',
          'MODEL':model_name+'.params',
          'VOCAB':model_name+'.vocabs',
          'EPOCH_MODEL':model_name+'_epoch.params'
    })


    # Preprocess
    vocabs = prepro.create_vocabs_from_trainset(trainset=hp['TRAINSET'],
                                                threshold=hp['THRESHOLD'],
                                                fn_dictionary=hp['DICTIONARY'],
                                                fn_vocabs=hp['VOCAB'],
                                                delimiter=delimiter,
                                                newline=newline)

    if emb_file is not None:
        embs, dim_word = prepro.embedding_loader(fn_embedding=hp['EMBEDDING'],
                                                 word2id=vocabs[2])
        hp['DIM_WORD'] = dim_word
    else:
        embs = None


    TrainData = prepro.from_file(filename=hp['TRAINSET'],
                                 window_size=hp['WINDOW_SIZE'],
                                 vocabs=vocabs,
                                 delimiter=delimiter,
                                 newline=newline)
    TestData  = prepro.from_file(filename=hp['TESTSET'],
                                 window_size=hp['WINDOW_SIZE'],
                                 vocabs=vocabs,
                                 delimiter=delimiter,
                                 newline=newline)
    DevData   = prepro.from_file(filename=hp['DEVSET'],
                                 window_size=hp['WINDOW_SIZE'],
                                 vocabs=vocabs,
                                 delimiter=delimiter,
                                 newline=newline)

    # Update hyper-parameters
    hp['NUM_TRAIN']         = len(TrainData.ws_data)
    hp['NUM_TEST']          = len(TestData.ws_data)
    hp['NUM_DEV']           = len(DevData.ws_data)
    hp['VOCAB_SIZE_UNI']    = len(vocabs[0])
    hp['VOCAB_SIZE_BI']     = len(vocabs[1])
    hp['VOCAB_SIZE_WORD']   = len(vocabs[2])
    hp['VOCAB_SIZE_POSTAG'] = len(vocabs[3])

    # Construct networks
    _model = model.Model(hp=hp, embs=embs)

    # Start training
    _start(hp, model=_model, train_data=TrainData, test_data=TestData, dev_data=DevData)


def _evaluation(hp, fn_model, data):
    tagger = Tagger(vocabs=hp['VOCAB'], params=fn_model, hp=hp['HYPERPARAMS'])

    def data_for_eval(words, postags):
        sent = []
        for w, p in zip(words, postags):
            p = w+"\t"+p
            if mecab_system_eval.PY_3 is True:
                w = w.encode("UTF-8")
                p = p.encode("UTF-8")
            sent.append([w, p])
        return sent

    sys_data = []
    ans_data = []
    indice   = [i for i in range(len(data.ws_data))]
    for i in indice:
        words   = data.words[i]
        pids    = data.pos_data[i][1]
        postags = [tagger.id2pos[pid] for pid in pids]
        ans_data.append(data_for_eval(words, postags))

        output      = tagger.tagging(''.join(words))
        sys_words   = output.words
        sys_postags = output.postags
        sys_data.append(data_for_eval(sys_words, sys_postags))

    r = mecab_system_eval.mecab_eval(sys_data, ans_data)
    _, _, ws_f, _, _, pos_f = mecab_system_eval.calculate_fvalues(r)
    return ws_f, pos_f


def _start(hp, model, train_data, test_data, dev_data):
    for k, v in hp.items():
        logging.info('[nagisa] {}: {}'.format(k, v))

    logs = '{:5}\t{:5}\t{:5}\t{:5}\t{:8}\t{:8}\t{:8}\t{:8}'.format(
        'Epoch', 'LR', 'Loss', 'Time_m', 'DevWS_f1',
        'DevPOS_f1', 'TestWS_f1', 'TestPOS_f1')
    logging.info(logs)


    utils.dump_data(hp, hp['HYPERPARAMS'])

    decay_counter  = 0
    best_dev_score = -1.0
    indice = [i for i in range(len(train_data.ws_data))]
    for e in range(1, hp['EPOCH']+1):
        t = time.time()
        losses = 0.
        random.shuffle(indice)
        for i in indice:
            # Word Segmentation
            X = train_data.ws_data[i][0]
            Y = train_data.ws_data[i][1]

            obs = model.encode_ws(X, train=True)
            gold_score = model.score_sentence(obs, Y)
            forward_score = model.forward(obs)
            loss = forward_score-gold_score
            # Update
            loss.backward()
            model.trainer.update()
            losses += loss.value()

            # POS-tagging
            X = train_data.pos_data[i][0]
            Y = train_data.pos_data[i][1]
            loss = model.get_POStagging_loss(X, Y)
            losses += loss.value()
            # Update
            loss.backward()
            model.trainer.update()

        model.model.save(hp['EPOCH_MODEL'])
        dev_ws_f, dev_pos_f = _evaluation(hp, fn_model=hp['EPOCH_MODEL'], data=dev_data)

        if dev_ws_f > best_dev_score:
            best_dev_score = dev_ws_f
            decay_counter = 0
            model.model.save(hp['MODEL'])
            test_ws_f, test_pos_f = _evaluation(hp, fn_model=hp['MODEL'], data=test_data)
        else:
            decay_counter += 1
            if decay_counter >= hp['DECAY']:
                model.trainer.learning_rate = model.trainer.learning_rate/2
                decay_counter = 0

        losses = losses/len(indice)
        logs = [e, model.trainer.learning_rate, losses, (time.time()-t)/60,
                dev_ws_f, dev_pos_f, test_ws_f, test_pos_f]

        logs = [log[:5] for log in map(str, logs)]

        logs = '{:5}\t{:5}\t{:5}\t{:5}\t{:8}\t{:8}\t{:8}\t{:8}'.format(
            logs[0], logs[1], logs[2], logs[3], logs[4],
            logs[5], logs[6], logs[7])

        logging.info(logs)




# -*- coding:utf-8 -*-

from __future__ import division, print_function, absolute_import

import numpy as np
import dynet_config
dynet_config.set(mem=32, random_seed=1234)
import dynet as dy


class Model(object):

    def __init__(self, hp, params=None, embs=None):
        # Set hyperparameters.
        dim_uni      = hp['DIM_UNI']
        dim_bi       = hp['DIM_BI']
        dim_ctype    = hp['DIM_CTYPE']
        dim_word     = hp['DIM_WORD']
        dim_tag_emb  = hp['DIM_TAGEMB']
        size_uni     = hp['VOCAB_SIZE_UNI']
        size_bi      = hp['VOCAB_SIZE_BI']
        size_word    = hp['VOCAB_SIZE_WORD']
        size_postags = hp['VOCAB_SIZE_POSTAG']
        lr           = hp['LEARNING_RATE']
        ws           = hp['WINDOW_SIZE']
        layers       = hp['LAYERS']
        dim_hidden   = hp['DIM_HIDDEN']
        self.dropout_rate = hp['DROPOUT_RATE']

        # Build a word segmentation model and a POS-tagging model simultaneously.
        model = dy.ParameterCollection()

        # Create embedding matrices.
        # These models share the unigram, bigramn and word embedding matrix.
        self.UNI   = model.add_lookup_parameters((size_uni, dim_uni))
        self.BI    = model.add_lookup_parameters((size_bi, dim_bi))
        if embs is not None:
            self.WORD  = model.add_lookup_parameters((size_word, dim_word), embs)
        else:
            self.WORD  = model.add_lookup_parameters((size_word, dim_word))
        self.CTYPE = model.add_lookup_parameters((7, dim_ctype))
        self.POS   = model.add_lookup_parameters((size_postags, dim_tag_emb))

        # Create a bi-directional LSTM model for word segmentation.
        dim_ws_input  = (dim_uni+dim_bi+dim_ctype)*ws+dim_word*2
        self.ws_model = dy.BiRNNBuilder(layers, dim_ws_input, dim_hidden, model, dy.LSTMBuilder)

        # Create a bi-directional LSTM model for POS-tagging.
        dim_pos_input  = dim_word+dim_uni+dim_tag_emb
        self.pos_model = dy.BiRNNBuilder(layers, dim_pos_input, dim_hidden, model, dy.LSTMBuilder)
        # Create a bi-directional LSTM model for encoding character sequence.
        self.char_seq_model = dy.BiRNNBuilder(layers, dim_uni, dim_uni, model, dy.LSTMBuilder)

        # Create trainer
        self.trainer = dy.SimpleSGDTrainer(model, lr)
        self.trainer.set_clip_threshold(5)

        # Set a tag scheme for WS model.
        dim_output = 4+2
        self.sp_s  = 4
        self.sp_e  = 5

        # Create weight matrices and bias vectors.
        self.dim_output = dim_output
        self.w_ws  = model.add_parameters((dim_output, dim_hidden))
        self.b_ws  = model.add_parameters(dim_output)
        self.trans = model.add_lookup_parameters((dim_output, dim_output))
        self.w_pos = model.add_parameters((size_postags, dim_hidden))
        self.b_pos = model.add_parameters(size_postags)
        self.dim_word    = dim_word
        self.dim_tag_emb = dim_tag_emb

        # Load trained parameters.
        if params:
            model.populate(params)
        self.model = model

        # As nparray
        self.trans_array = self.trans.as_array()


    def encode_ws(self, X, train=False):
        dy.renew_cg()

        # Remove dy.parameters(...) for DyNet v.2.1
        #w_ws = dy.parameter(self.w_ws)
        #b_ws = dy.parameter(self.b_ws)
        w_ws = self.w_ws
        b_ws = self.b_ws

        ipts = []
        length = len(X[0])
        for i in range(length):
            uni   = X[0][i]
            bi    = X[1][i]
            ctype = X[2][i]
            start = X[3][i]
            end   = X[4][i]

            vec_uni   = dy.concatenate([self.UNI[uid] for uid in uni])
            vec_bi    = dy.concatenate([self.BI[bid] for bid in bi])
            vec_start = dy.esum([self.WORD[sid] for sid in start])
            vec_end   = dy.esum([self.WORD[eid] for eid in end])
            vec_ctype = dy.concatenate([self.CTYPE[cid] for cid in ctype])
            vec_at_i  = dy.concatenate([vec_uni, vec_bi, vec_ctype, vec_start, vec_end])

            if train is True:
                vec_at_i = dy.dropout(vec_at_i, self.dropout_rate)
            ipts.append(vec_at_i)

        bilstm_outputs = self.ws_model.transduce(ipts)
        observations   = [w_ws*h+b_ws for h in bilstm_outputs]
        return observations


    def forward(self, observations):

        def log_sum_exp(scores):
            npval = scores.npvalue()
            argmax_score = np.argmax(npval)
            max_score_expr = dy.pick(scores, argmax_score)
            max_score_expr_broadcast = dy.concatenate([max_score_expr] * self.dim_output)
            return max_score_expr + dy.log(dy.sum_elems(dy.transpose(dy.exp(scores - max_score_expr_broadcast))))

        init_alphas = [-1e10] * self.dim_output
        init_alphas[self.sp_s] = 0
        for_expr = dy.inputVector(init_alphas)
        for obs in observations:
            alphas_t = []
            for next_tag in range(self.dim_output):
                obs_broadcast = dy.concatenate([dy.pick(obs, next_tag)] * self.dim_output)
                next_tag_expr = for_expr + self.trans[next_tag] + obs_broadcast
                alphas_t.append(log_sum_exp(next_tag_expr))
            for_expr = dy.concatenate(alphas_t)
        terminal_expr = for_expr + self.trans[self.sp_e]
        alpha = log_sum_exp(terminal_expr)
        return alpha


    def score_sentence(self, observations, tags):
        if not len(observations) == len(tags):
            raise AssertionError("len(observations) != len(tags)")

        score_seq = [0]
        score = dy.scalarInput(0)
        tags = [self.sp_s] + tags
        for i, obs in enumerate(observations):
            score = score + dy.pick(self.trans[tags[i+1]], tags[i]) + dy.pick(obs, tags[i+1])
            score_seq.append(score.value())
        score = score + dy.pick(self.trans[self.sp_e], tags[-1])
        return score

    # Nagisa does not use this method.
    #
    # def viterbi_decoding(self, observations):
    #     backpointers = []
    #     init_vvars   = [-1e10] * self.dim_output
    #     init_vvars[self.sp_s] = 0
    #     for_expr     = dy.inputVector(init_vvars)
    #     trans_exprs  = [self.trans[idx] for idx in range(self.dim_output)]
    #     for obs in observations:
    #         bptrs_t = []
    #         vvars_t = []
    #         for next_tag in range(self.dim_output):
    #             next_tag_expr = for_expr + trans_exprs[next_tag]
    #             next_tag_arr = next_tag_expr.npvalue()
    #             best_tag_id  = np.argmax(next_tag_arr)
    #             bptrs_t.append(best_tag_id)
    #             vvars_t.append(dy.pick(next_tag_expr, best_tag_id))
    #         for_expr = dy.concatenate(vvars_t) + obs
    #         backpointers.append(bptrs_t)
    #     terminal_expr = for_expr + trans_exprs[self.sp_e]
    #     terminal_arr  = terminal_expr.npvalue()
    #     best_tag_id   = np.argmax(terminal_arr)
    #     path_score    = dy.pick(terminal_expr, best_tag_id)
    #     best_path = [best_tag_id]
    #     for bptrs_t in reversed(backpointers):
    #         best_tag_id = bptrs_t[best_tag_id]
    #         best_path.append(best_tag_id)
    #     start = best_path.pop()
    #     best_path.reverse()
    #     if not start == self.sp_s:
    #         raise AssertionError("start != self.sp_s")
    #     return best_path, path_score

    def encode_pt(self, X, train=False):
        dy.renew_cg()

        # Remove dy.parameters(...) for DyNet v.2.1
        #w_pos = dy.parameter(self.w_pos)
        #b_pos = dy.parameter(self.b_pos)
        w_pos = self.w_pos
        b_pos = self.b_pos

        ipts  = []
        length = len(X[0])
        for i in range(length):
            cids = X[0][i]
            wid  = X[1][i]
            tids = X[2][i]
            vec_char = self.char_seq_model.transduce([self.UNI[cid] for cid in cids])[-1]

            vec_tags = []
            for tid in tids:
                if tid == 0:
                    zero = dy.inputVector(np.zeros(self.dim_tag_emb))
                    vec_tags.append(zero)
                else:
                    vec_tags.append(self.POS[tid])
            vec_tag = dy.esum(vec_tags)

            if wid == 0:
                vec_word = dy.inputVector(np.zeros(self.dim_word))
            else:
                vec_word = self.WORD[wid]

            vec_at_i = dy.concatenate([vec_word, vec_char, vec_tag])
            if train is True:
                vec_at_i = dy.dropout(vec_at_i, self.dropout_rate)
            ipts.append(vec_at_i)
        hiddens = self.pos_model.transduce(ipts)
        probs = [dy.softmax(w_pos*h+b_pos) for h in hiddens]
        return probs


    def get_POStagging_loss(self, X, Y):
        losses = []
        probs = self.encode_pt(X, train=True)
        for prob, y in zip(probs, Y):
            losses.append(-dy.log(dy.pick(prob, y)))
        loss = dy.esum(losses)
        return loss


    def POStagging(self, X):
        probs = self.encode_pt(X)
        pids = [np.argmax(prob.npvalue()) for prob in probs]
        return pids



import nagisa_utils as utils

from nagisa.tagger import Tagger
from nagisa.train import fit

version = '0.2.8'
# Initialize instance
tagger  = Tagger()
# Functions
wakati  = tagger.wakati
tagging = tagger.tagging
filter  = tagger.filter
extract = tagger.extract
postagging = tagger.postagging
decode = tagger.decode
fit = fit

__version__ = version



# -*- coding:utf-8 -*-

import io
import os
import sys

from setuptools import setup
from setuptools.extension import Extension

readme = 'README.md'

try:
    from pypandoc import convert_file
    long_description = convert_file(readme, 'rst')
except ImportError:
    if os.name == 'nt':
        if sys.version_info.major == 2:
            f = io.open(readme, 'r', encoding='utf_8_sig')
        else:
            f = open(readme, 'r', encoding='utf_8_sig')
    else:
        f = open(readme, 'r')
    long_description = f.read()


classifiers = [
    'License :: OSI Approved :: MIT License',
    'Natural Language :: Japanese',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Operating System :: Unix',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
    'Topic :: Text Processing :: Linguistic',
    'Topic :: Software Development :: Libraries :: Python Modules'
]


class defer_cythonize(list):
    def __init__(self, callback):
        self._list, self.callback = None, callback

    def c_list(self):
        if self._list is None:
            self._list = self.callback()
        return self._list

    def __iter__(self):
        for elem in self.c_list():
            yield elem

    def __getitem__(self, ii):
        return self.c_list()[ii]

    def __len__(self):
        return len(self.c_list())


def extensions():
    from Cython.Build import cythonize
    import numpy
    extensions = [Extension('nagisa_utils',
                  ['nagisa/nagisa_utils.pyx'],
                  include_dirs=[numpy.get_include()])]
    return cythonize(extensions)


def switch_install_requires():
    major = sys.version_info.major
    minor = sys.version_info.minor
    if os.name == 'posix' and major == 3 and minor > 7:
        return ['six', 'numpy', 'DyNet38']
    else:
        return ['six', 'numpy', 'DyNet']


setup(
    name='nagisa',
    packages=['nagisa'],
    author='Taishi Ikeda',
    author_email='taishi.ikeda.0323@gmail.com',
    version='0.2.8',
    description='A Japanese tokenizer based on recurrent neural networks',
    long_description_content_type="text/markdown",
    long_description=long_description,
    url='https://github.com/taishi-i/nagisa',
    download_url='https://github.com/taishi-i/nagisa/archive/0.2.8.tar.gz',
    license='MIT License',
    platforms='Unix',
    setup_requires=['six', 'cython', 'numpy'],
    install_requires=switch_install_requires(),
    classifiers=classifiers,
    include_package_data=True,
    test_suite='test.nagisa_test.suite',
    ext_modules=defer_cythonize(extensions)
)



# -*- coding:utf-8 -*-

from __future__ import division, print_function, absolute_import

import sys
import codecs

if sys.version_info.major == 3:
    PY_3 = True
else:
    PY_3 = False


def readFile(filename):
    sent = []
    data = []
    with codecs.open(filename, 'r', encoding='utf_8_sig') as f:
        for line in f:
            line = line.rstrip()
            if line == "EOS":
                data.append(sent)
                sent = []
            else:
                surface, csv_form = line.split("\t")
                csv_form = line
                if PY_3 is True:
                    surface  = surface.encode("UTF-8")
                    csv_form = csv_form.encode("UTF-8")
                sent.append([surface, csv_form])
    return data


def mecab_eval(sys_data, ans_data):
    """
    This script is written by referring to the following code.
    https://github.com/taku910/mecab/blob/master/mecab/src/eval.cpp
    """
    if not len(sys_data) == len(ans_data):
        raise AssertionError("len(sys_data) != len(ans_data)")
    num_sents = len(sys_data)

    prec = 0
    recall = 0
    num_correct = [0, 0]
    for i in range(num_sents):
        i_sys = 0
        i_ans = 0
        l_sys = 0
        l_ans = 0
        sys_sent = sys_data[i]
        ans_sent = ans_data[i]

        while ((i_sys < len(sys_sent)) & (i_ans < len(ans_sent))):
            if (l_sys == l_ans):
                if (sys_sent[i_sys][0] == ans_sent[i_ans][0]):
                    num_correct[0] += 1 # Word Segmentation

                if (sys_sent[i_sys][1] == ans_sent[i_ans][1]):
                    num_correct[1] += 1 # POS-tagging

                l_sys  += len(sys_sent[i_sys][0])
                l_ans  += len(ans_sent[i_ans][0])
                i_sys  += 1
                i_ans  += 1
                prec   += 1
                recall += 1
            elif (l_sys < l_ans):
                l_sys  += len(sys_sent[i_sys][0])
                i_sys  += 1
                prec   += 1
            else:
                l_ans  += len(ans_sent[i_ans][0])
                i_ans  += 1
                recall += 1

        while (i_sys < len(sys_sent)):
            i_sys  += 1
            prec   += 1

        while (i_ans < len(ans_sent)):
            i_ans  += 1
            recall += 1

    ws_c = num_correct[0]
    pt_c = num_correct[1]
    return [ws_c, pt_c, prec, recall]


def calculate_fvalues(r):
    ws_c, pt_c, prec, recall = r

    ws_p = round(100*ws_c/prec, 4)
    ws_r = round(100*ws_c/recall, 4)
    if ws_p+ws_r == 0:
        ws_f = round(0., 4)
    else:
        ws_f = round(2*ws_p*ws_r/(ws_p+ws_r), 4)

    pt_p = round(100*pt_c/prec, 4)
    pt_r = round(100*pt_c/recall, 4)
    if pt_p+pt_r == 0:
        pt_f = round(0., 4)
    else:
        pt_f = round(2*pt_p*pt_r/(pt_p+pt_r), 4)

    return [ws_p, ws_r, ws_f, pt_p, pt_r, pt_f]


def print_eval(r):
    ws_c, pt_c, prec, recall = r
    ws_p, ws_r, ws_f, pt_p, pt_r, pt_f = calculate_fvalues(r)

    ws_p_out = str(ws_p)+"("+str(ws_c)+"/"+str(prec)+")"
    ws_r_out = str(ws_r)+"("+str(ws_c)+"/"+str(recall)+")"
    pt_p_out = str(pt_p)+"("+str(pt_c)+"/"+str(prec)+")"
    pt_r_out = str(pt_r)+"("+str(pt_c)+"/"+str(recall)+")"

    ws_out = ["LEVEL 0:", ws_p_out, ws_r_out, str(ws_f)]
    pt_out = ["LEVEL ALL:", pt_p_out, pt_r_out, str(pt_f)]

    header = ["        ", "precision", "recall", "F"]
    print("\t".join(header))
    print("\t".join(ws_out))
    print("\t".join(pt_out))


if __name__ == "__main__":
    import argparse

    dsc = "Reimplementation of mecab-system-eval \
           (https://github.com/taku910/mecab/blob/master/mecab/src/eval.cpp). \
           However, this script only evaluates on the word segmentation level \
           and POS-tagging level. Input file format: Word\tPOS-tag"

    parser = argparse.ArgumentParser(description=dsc)
    parser.add_argument("--system", type=str, help="System output's file")
    parser.add_argument("--answer", type=str, help="Answer file")
    args = parser.parse_args()

    sys_data = readFile(args.system)
    ans_data = readFile(args.answer)

    r = mecab_eval(sys_data, ans_data)
    print_eval(r)
